import java.awt.*;
import java.util.*;
import java.lang.*;
import OpticsObject;
import OpticsObjPanel;
import LightRay;
import HitObject;
import HitGlass;

import java.lang.Math;

class CircularLensObject extends GlassObject
{
  protected static final double MAX_LENGTH = 10.0;

  double fLength = 5.0;
  double fCircRadius = 5.0;

  double fCircX;
  double fCircY;
  double fBeginAngle;
  double fEndAngle;

  static int nObjectNo = 1;
  public String GetUniqueName () { return "Circular Lens "+(nObjectNo++); }
  public static String GetPanelName () { return "CircularLensPanel"; }
  public static String GetObjectName () { return "Circular Lens"; }

  public void SetPanel (OpticsObjPanel panel)
  {
    super.SetPanel (panel);
    if (thePanel != null)
      {
	((CircularLensPanel) thePanel).SetLength (fLength);
	((CircularLensPanel) thePanel).SetRadius (fCircRadius);
      }
  }

  public void SetLength (double length,boolean setPanel)
  {
    theLabCanvas.AbortCalculation ();
    if (fCircRadius == fLength)
      fCircRadius = length;
    fLength = length;
    if (fLength > fCircRadius)
      fCircRadius = fLength;
    fRadius = fLength;
    nLabRadius = (int) (fLength * 10.0)+3;
    if (setPanel && thePanel != null)
      {
	((CircularLensPanel) thePanel).SetLength (fLength);
	((CircularLensPanel) thePanel).SetRadius (fCircRadius);
      }
    UpdateInternalState ();
    theLabCanvas.repaint ();
    theLabCanvas.Recalculate ();
  }

  public void SetRadius (double radius,boolean setPanel)
  {
    theLabCanvas.AbortCalculation ();
    fCircRadius = (radius > fLength ? radius : fLength);
    if (setPanel && thePanel != null)
      ((CircularLensPanel) thePanel).SetRadius (fCircRadius);
    UpdateInternalState ();
    theLabCanvas.repaint ();
    theLabCanvas.Recalculate ();
  }

  protected void UpdateInternalState()
    {
      double fCenterCircDist = Math.sqrt (fCircRadius*fCircRadius-
					  fLength*fLength);
      
      fCircX = fX - fCenterCircDist * Math.cos (fDirection);
      fCircY = fY + fCenterCircDist * Math.sin (fDirection);
      
      fBeginAngle = Math.atan2 (fCenterCircDist * Math.sin (fDirection)-
				fLength * Math.cos (fDirection),
				fCenterCircDist * Math.cos (fDirection)+
				fLength * Math.sin (fDirection));
      fEndAngle = Math.atan2 (fCenterCircDist * Math.sin (fDirection)+
			      fLength * Math.cos (fDirection),
			      fCenterCircDist * Math.cos (fDirection)-
			      fLength * Math.sin (fDirection));
      
      if (fCircRadius < 0)
	{
	  double f = fBeginAngle;
	  fBeginAngle = fEndAngle;
	  fEndAngle = f;
	}
      
      if (fEndAngle < fBeginAngle)
	fEndAngle += 2*Math.PI;
      
      //    System.out.println ("fCircX: "+fCircX);
      //    System.out.println ("fCircY: "+fCircY);
      //    System.out.println ("fBeginAngle: "+fBeginAngle);
      //    System.out.println ("fEndAngle: "+fEndAngle);
    }
  
  public void DrawContours (Graphics g,boolean selected)
    {
      g = InitDrawContours (g,selected);
      
      double fXLength =  fLength*10.0*Math.cos (0.5*Math.PI+fDirection);
      double fYLength = -fLength*10.0*Math.sin (0.5*Math.PI+fDirection);
      
      int circX = (int) ((fCircX-fX)*10.0);
      int circY = (int) ((fCircY-fY)*10.0);
      
      int circRadius = (int) (Math.abs(fCircRadius)*10.0);
      
      g.drawLine ((int) fXLength,(int) fYLength,
		  (int) -fXLength,(int) -fYLength);
      
      g.drawArc (circX-circRadius,circY-circRadius,
		 2*circRadius,2*circRadius,
		 RadiansToIntDegrees (fBeginAngle),
		 RadiansToIntDegrees(ToAboveZeroInterval(fEndAngle-
							 fBeginAngle)));
      
    }
  
  public void HandlePointMove (int nPoint,Point pt)
  {
    super.HandlePointMove (nPoint,pt);

    if (nPoint == 1)
      {
	double fDirection = GetPointDirection (pt);
	double fLength = GetPointLength (pt);

	if (fLength > MAX_LENGTH)
	  fLength = MAX_LENGTH;
	SetDirection (ToPiInterval (fDirection-0.5*Math.PI),true);
	SetLength (fLength,true);
      }
    if (nPoint == 2)
      {
	double radius = GetPointLength (pt);

	if (radius > fLength)
	  {
	    radius = fLength;
	  }
	else
	  {
	    double x3 = GetPointX(pt);
	    double y3 = GetPointY(pt);
	    
	    double x1 = fLength * Math.sin (fDirection);
	    double y1 = fLength * Math.cos (fDirection);
	    
	    double x2 = -fLength * Math.sin (fDirection);
	    double y2 = -fLength * Math.cos (fDirection);
	    
	    double t = 0.5*(((y2-y1)*(y1-y3)-(x2-x1)*(x3-x1))/
			    ((y3-y2)*(x3-x1)-(x2-x3)*(y1-y3)));
	    
	    double dx = (x2-x3)/2+t*(y3-y2);
	    double dy = (y2-y3)/2+t*(x2-x3);
	    
	    radius = Math.sqrt(dx*dx+dy*dy);
	  }

	SetRadius (radius,true);
      }
  }
  
  public Vector GetHandlePoints ()
    {
      Vector v = super.GetHandlePoints ();
      
      v.addElement (new Point ((int) (( fLength*10.0*
					Math.cos (0.5*Math.PI+fDirection))),
			       (int) ((-fLength*10.0*
				       Math.sin (0.5*Math.PI+fDirection)))));
      
      double fCenterCircDist = fCircRadius +
	(fCircRadius > 0 ? 
	 (-Math.sqrt (fCircRadius*fCircRadius-fLength*fLength)) :
	 (Math.sqrt (fCircRadius*fCircRadius-fLength*fLength)));
      v.addElement (new Point ((int) ((-fCenterCircDist*10.0*
				       Math.cos (Math.PI+fDirection))),
			     (int) ((fCenterCircDist*10.0*
				     Math.sin (Math.PI+fDirection)))));
      return v;
    }
  
  public double IsHit (LightRay ray)
    {
      double t1 = IntersectLine (ray,
				 fX + fLength*Math.cos (0.5*Math.PI+fDirection),
				 fY - fLength*Math.sin (0.5*Math.PI+fDirection),
				 (- Math.cos (0.5*Math.PI+fDirection)),
				 (+ Math.sin (0.5*Math.PI+fDirection)),
				 fLength,1);
      
      double t2 = IntersectCircle (ray,
				   fCircX,fCircY,
				   fCircRadius,
				   fBeginAngle,fEndAngle,2);

      return Math.min (t1,t2);
    }
  
  public HitObject GetHit (LightRay ray)
    {
      double t1 = IntersectLine (ray,
				 fX + fLength*Math.cos (0.5*Math.PI+fDirection),
				 fY - fLength*Math.sin (0.5*Math.PI+fDirection),
				 (- Math.cos (0.5*Math.PI+fDirection)),
				 (+ Math.sin (0.5*Math.PI+fDirection)),
				 fLength,1);
      
      double t2 = IntersectCircle (ray,
				   fCircX,fCircY,
				   fCircRadius,
				   fBeginAngle,fEndAngle,1);

      if (t1 < t2)
	return new HitGlass (IntersectHitLine (ray,
					       fX + fLength*
					       Math.cos (0.5*Math.PI+fDirection),
					       fY - fLength*
					       Math.sin (0.5*Math.PI+fDirection),
					       (- Math.cos (0.5*Math.PI+fDirection)),
					       (+ Math.sin (0.5*Math.PI+fDirection)),
					       1,Math.PI-fDirection),
			     GetRefractionConsts ());
      else
	return new HitGlass (IntersectHitCircle (ray,
						 fCircX,fCircY,
						 fCircRadius,
						 fBeginAngle,fEndAngle,
						 2,true),
			     GetRefractionConsts ());
    }
}
