import java.awt.*;
import java.util.*;
import java.lang.*;
import GlassObject;
import GlassPanel;
import LightRay;
import HitObject;
import HitGlass;

import java.lang.Math;

class EquilateralTrianglePrismObject extends GlassObject
{
  protected static final double MAX_LENGTH = 10.0;

  double fLength = 5.0;

  static int nObjectNo = 1;
  public String GetUniqueName () { return "Eq. Tri. Prism "+(nObjectNo++); }
  public static String GetPanelName () {return"EquilateralTrianglePrismPanel";}
  public static String GetObjectName () {return "Equilateral Triangle Prism"; }

  public void SetPanel (OpticsObjPanel panel)
  {
    super.SetPanel (panel);
    if (thePanel != null)
      {
	((EquilateralTrianglePrismPanel) thePanel).SetLength (fLength);
      }
    fRadius = fLength*2/Math.sqrt(3);
  }

  public void SetLength (double length,boolean setPanel)
  {
    theLabCanvas.AbortCalculation ();
    fLength = length;
    fRadius = fLength*2/Math.sqrt(3);
    nLabRadius = (int) (fRadius * 10.0);
    if (setPanel && thePanel != null)
      ((EquilateralTrianglePrismPanel) thePanel).SetLength (fLength);
    theLabCanvas.repaint ();
    theLabCanvas.Recalculate ();
  }

  public void DrawContours (Graphics g,boolean selected)
  {
    g = InitDrawContours (g,selected);

    int x1 = (int) ( fRadius*10.0*Math.cos (0.5*Math.PI+fDirection));
    int y1 = (int) (-fRadius*10.0*Math.sin (0.5*Math.PI+fDirection));
    int x2 = (int) ( fRadius*10.0*Math.cos ((0.5+2.0/3.0)*Math.PI+fDirection));
    int y2 = (int) (-fRadius*10.0*Math.sin ((0.5+2.0/3.0)*Math.PI+fDirection));
    int x3 = (int) ( fRadius*10.0*Math.cos ((0.5-2.0/3.0)*Math.PI+fDirection));
    int y3 = (int) (-fRadius*10.0*Math.sin ((0.5-2.0/3.0)*Math.PI+fDirection));

    g.drawLine (x1,y1,x2,y2);
    g.drawLine (x2,y2,x3,y3);
    g.drawLine (x3,y3,x1,y1);
  }

  public void HandlePointMove (int nPoint,Point pt)
  {
    super.HandlePointMove (nPoint,pt);

    if (nPoint == 1)
      {
	double direction = GetPointDirection (pt);
	double length = GetPointLength (pt)*Math.sqrt(3)/2;

	if (length > MAX_LENGTH)
	  length = MAX_LENGTH;
	SetDirection (ToPiInterval (direction-0.5*Math.PI),true);
	SetLength (length,true);
      }
  }
  
  public Vector GetHandlePoints ()
  {
    Vector v = super.GetHandlePoints ();
    
    v.addElement (new Point ((int) (( fRadius*10.0*
				     Math.cos (0.5*Math.PI+fDirection))),
			     (int) ((-fRadius*10.0*
				     Math.sin (0.5*Math.PI+fDirection)))));
    return v;
  }

  public double IsHit (LightRay ray)
  {
    //    if (ray.lastHitObject == this)
    //      return 0.0;

    double t = Double.MAX_VALUE;

    for (int i = 0; i < 3; i++)
      if (ray.lastHitObject != this || ray.lastHitSurface != i)
	{
	  double x1 = fX + fRadius*Math.cos ((0.5+i*(2.0/3.0))*Math.PI+
					     fDirection);
	  double y1 = fY - fRadius*Math.sin ((0.5+i*(2.0/3.0))*Math.PI+
					     fDirection);
	  double vx = - Math.cos ((0.5+i*(2.0/3.0)-1.0/6.0)*Math.PI+
				  fDirection);
	  double vy = + Math.sin ((0.5+i*(2.0/3.0)-1.0/6.0)*Math.PI+
				  fDirection);
	  
	  double t1 = (vx*(ray.pY-y1)-vy*(ray.pX-x1))/(ray.vX*vy-ray.vY*vx);
	  if (t1 < 0.0 || t1 > t)
	    continue;
	  double s = (Math.abs(vx) > Math.abs(vy) ? 
		      (ray.pX+t1*ray.vX-x1)/vx : 
		      (ray.pY+t1*ray.vY-y1)/vy);
	  if (s < 0.0 || s > 2*fLength)
	    continue;
	  t = t1;	  
	}
    return t;
  }

  public HitObject GetHit (LightRay ray)
  {
    double t = Double.MAX_VALUE;
    int nSurf = 0;

    for (int i = 0; i < 3; i++)
      if (ray.lastHitObject != this || ray.lastHitSurface != i)
	{
	  double x1 = fX + fRadius*Math.cos ((0.5+i*(2.0/3.0))*Math.PI+
					     fDirection);
	  double y1 = fY - fRadius*Math.sin ((0.5+i*(2.0/3.0))*Math.PI+
					     fDirection);
	  double vx = - Math.cos ((0.5+i*(2.0/3.0)-1.0/6.0)*Math.PI+
				  fDirection);
	  double vy = + Math.sin ((0.5+i*(2.0/3.0)-1.0/6.0)*Math.PI+
				  fDirection);
	  
	  double t1 = (vx*(ray.pY-y1)-vy*(ray.pX-x1))/(ray.vX*vy-ray.vY*vx);
	  if (t1 < 0.0 || t1 > t)
	    continue;
	  double s = (Math.abs(vx) > Math.abs(vy) ? 
		      (ray.pX+t1*ray.vX-x1)/vx : 
		      (ray.pY+t1*ray.vY-y1)/vy);
	  if (s < 0.0 || s > 2*fLength)
	    continue;
	  t = t1;	  
	  nSurf = i;
	}
    return new HitGlass (ray.pX+t*ray.vX,ray.pY+t*ray.vY,
			 nSurf,Math.PI-((nSurf*(2.0/3.0)-1.0/6.0)*Math.PI+
					fDirection),GetRefractionConsts ());
  }
}
