import java.awt.*;
import java.util.*;
import java.lang.*;
import OpticsObject;
import OpticsObjPanel;
import LightRay;
import HitObject;
import HitGlass;

import java.lang.Math;

class GeneralLensObject extends GlassObject
{
  protected static final double MAX_LENGTH = 10.0;
  protected static final double MAX_WIDTH = 15.0;

  double fLength = 5.0;
  double fWidth = 2.0;
  double fCircRadius[] = { 50.0, 50.0 };

  double fCircX[] = new double[2];
  double fCircY[] = new double[2];
  double fBeginAngle[] = new double[2];
  double fEndAngle[] = new double[2];

  static int nObjectNo = 1;
  public String GetUniqueName () { return "General Lens "+(nObjectNo++); }
  public static String GetPanelName () { return "GeneralLensPanel"; }
  public static String GetObjectName () { return "General Lens"; }

  public void SetPanel (OpticsObjPanel panel)
  {
    super.SetPanel (panel);
    if (thePanel != null)
      {
	((GeneralLensPanel) thePanel).SetLength (fLength);
	((GeneralLensPanel) thePanel).SetWidth (fWidth);
	for (int i = 0; i < 2; i++)
	  ((GeneralLensPanel) thePanel).SetRadius (fCircRadius[i],i);
      }
  }

  public void SetLength (double length,boolean setPanel)
  {
    theLabCanvas.AbortCalculation ();
    for (int i = 0; i < 2; i++)
      {
	if (fCircRadius[i] == fLength)
	  fCircRadius[i] = length;
	else if (fCircRadius[i] == -fLength)
	  fCircRadius[i] = -length;
      }
    fLength = length;
    for (int i = 0; i < 2; i++)
      if (fLength > Math.abs(fCircRadius[i]))
	fCircRadius[i] = fCircRadius[i] > 0 ? fLength : -fLength;
    if (setPanel && thePanel != null)
      {
	((GeneralLensPanel) thePanel).SetLength (fLength);
	for (int i = 0; i < 2; i++)
	  ((GeneralLensPanel) thePanel).SetRadius (fCircRadius[i],i);
      }
    UpdateInternalState ();
    theLabCanvas.repaint ();
    theLabCanvas.Recalculate ();
  }

  public void SetWidth (double width,boolean setPanel)
  {
    theLabCanvas.AbortCalculation ();
    fWidth = width;
    if (setPanel && thePanel != null)
      ((GeneralLensPanel) thePanel).SetWidth (fWidth);
    UpdateInternalState ();
    theLabCanvas.repaint ();
    theLabCanvas.Recalculate ();
  }

  public void SetRadius (double radius,int no,boolean setPanel)
  {
    theLabCanvas.AbortCalculation ();
    fCircRadius[no] = (Math.abs(radius) > fLength ?
		       radius : (radius > 0 ? fLength : -fLength));
    if (setPanel && thePanel != null)
      ((GeneralLensPanel) thePanel).SetRadius (fCircRadius[no],no);
    UpdateInternalState ();
    theLabCanvas.repaint ();
    theLabCanvas.Recalculate ();
  }

  protected void UpdateInternalState()
    {
	fRadius = Math.sqrt(fLength*fLength+fWidth*fWidth);
      for (int i = 0; i < 2; i++)
	{
	  double fCenterCircDist = -fWidth +
		(fCircRadius[i] < 0 ? 
		 (-Math.sqrt (fCircRadius[i]*fCircRadius[i]-fLength*fLength)) :
		 (Math.sqrt (fCircRadius[i]*fCircRadius[i]-fLength*fLength)));

	  if (Math.abs(fCenterCircDist-fCircRadius[i]) > fRadius)
	      fRadius = Math.abs(fCenterCircDist-fCircRadius[i]);
	  
	  double fDir = fDirection;
	  if (i == 1)
	    fDir += Math.PI;
	  
	  fCircX[i] = fX - (fCenterCircDist) * Math.cos (fDir);
	  fCircY[i] = fY + (fCenterCircDist) * Math.sin (fDir);
	  
	  fBeginAngle[i]  = Math.atan2 (fCenterCircDist * Math.sin (fDir)-fLength * Math.cos (fDir)-
					                                  fWidth * Math.cos (Math.PI*.5+fDir),
				       fCenterCircDist * Math.cos (fDir)+fLength * Math.sin (fDir)+
					                                 fWidth * Math.sin (Math.PI*.5+fDir));
	  fEndAngle[i] = Math.atan2 (fCenterCircDist * Math.sin (fDir)+fLength * Math.cos (fDir)-
					                               fWidth * Math.cos (Math.PI*.5+fDir),
				     fCenterCircDist * Math.cos (fDir)-fLength * Math.sin (fDir)+
					                               fWidth * Math.sin (Math.PI*.5+fDir));
	  
	  if (fCircRadius[i] < 0)
	    {
	      double f = fBeginAngle[i];
	      fBeginAngle[i] = fEndAngle[i];
	      fEndAngle[i] = f;
	    }
	  
	  if (fEndAngle[i] < fBeginAngle[i])
	    fEndAngle[i] += 2*Math.PI;
	}
      nLabRadius = (int) (fRadius * 10.0)+3;
	  
      //    System.out.println ("fCircX: "+fCircX);
      //    System.out.println ("fCircY: "+fCircY);
      //    System.out.println ("fBeginAngle: "+fBeginAngle);
      //    System.out.println ("fEndAngle: "+fEndAngle);
    }
  
  public void DrawContours (Graphics g,boolean selected)
    {
      g = InitDrawContours (g,selected);
      
      double fXLength =  fLength*10.0*Math.cos (0.5*Math.PI+fDirection);
      double fYLength = -fLength*10.0*Math.sin (0.5*Math.PI+fDirection);

      double fXWidth =  fWidth*10.0*Math.cos (fDirection);
      double fYWidth = -fWidth*10.0*Math.sin (fDirection);

      for (int i = 0; i < 2; i++)
	{
	  int circX = (int) ((fCircX[i]-fX)*10.0);
	  int circY = (int) ((fCircY[i]-fY)*10.0);
	  
	  int circRadius = (int) (Math.abs(fCircRadius[i])*10.0);
	  
	  g.drawArc (circX-circRadius,circY-circRadius,
		     2*circRadius,2*circRadius,
		     RadiansToIntDegrees (fBeginAngle[i]),
		     RadiansToIntDegrees(ToAboveZeroInterval(fEndAngle[i]-
							     fBeginAngle[i])));
	}

      g.drawLine ((int) ( fXLength+fXWidth),(int) ( fYLength+fYWidth),
      		  (int) ( fXLength-fXWidth),(int) ( fYLength-fYWidth));
      g.drawLine ((int) (-fXLength+fXWidth),(int) (-fYLength+fYWidth),
      		  (int) (-fXLength-fXWidth),(int) (-fYLength-fYWidth));
    }
  
  public void HandlePointMove (int nPoint,Point pt)
  {
    super.HandlePointMove (nPoint,pt);

    if (nPoint == 1)
      {
	double fDirection = GetPointDirection (pt);
	double fLength = GetPointLength (pt);

	if (fLength > MAX_LENGTH)
	  fLength = MAX_LENGTH;
	SetDirection (ToPiInterval (fDirection-0.5*Math.PI),true);
	SetLength (fLength,true);
      }
    else if (nPoint == 2)
      {
	  double fWidth = GetPointLength (pt,
					    fLength*Math.cos (0.5*Math.PI+fDirection),
					    -fLength*Math.sin (0.5*Math.PI+fDirection));
	  
	  if (fWidth > MAX_LENGTH)
	      fWidth = MAX_LENGTH;
	  SetWidth (fWidth,true);
      }
    else
      for (int i = 0; i < 2; i++)
	if (nPoint == 3+i)
	  {
	    double d = GetPointLengthDir (pt,(i==0?1:-1)*Math.cos (fDirection),
					  (i==0?1:-1)*-Math.sin (fDirection));
	    double radius = fLength;
	    
	    if (d < fLength+fWidth)
	      {
		  radius = (d*d+fLength*fLength+fWidth*fWidth-2*d*fWidth)/(2*(d-fWidth));
	      }
	    
	    SetRadius (radius,i,true);
	  }
  }
  
  public Vector GetHandlePoints ()
    {
      Vector v = super.GetHandlePoints ();
      
      v.addElement (new Point ((int) ( fLength*10.0*Math.cos (0.5*Math.PI+fDirection)),
			       (int) (-fLength*10.0*Math.sin (0.5*Math.PI+fDirection))));
      
      v.addElement (new Point ((int) ( fLength*10.0*Math.cos (0.5*Math.PI+fDirection)+
				       fWidth*10.0*Math.cos (fDirection)),
			       (int) (-fLength*10.0*Math.sin (0.5*Math.PI+fDirection)+
				      -fWidth*10.0*Math.sin (fDirection))));
      
      for (int i = 0; i < 2; i++)
	{
	    double fCenterCircDist = -fWidth - fCircRadius[i] +
		(fCircRadius[i] < 0 ? 
		 (-Math.sqrt (fCircRadius[i]*fCircRadius[i]-fLength*fLength)) :
		 (Math.sqrt (fCircRadius[i]*fCircRadius[i]-fLength*fLength)));
	    
	    v.addElement (new Point ((int) ((-fCenterCircDist*10.0*
					     (i==0?-1:1)*Math.cos (Math.PI+fDirection))),
				     (int) ((fCenterCircDist*10.0*
					     (i==0?-1:1)*Math.sin (Math.PI+fDirection)))));
	}
      return v;
    }
  
  public double IsHit (LightRay ray)
    {
      double t1 = IntersectLine (ray,
				 fX + fLength *Math.cos (0.5*Math.PI+fDirection)+
				      fWidth*Math.cos (fDirection),
				 fY - fLength*Math.sin (0.5*Math.PI+fDirection)-
				      fWidth*Math.sin (fDirection),
				 (- Math.cos (fDirection)),
				 (+ Math.sin (fDirection)),
				 fWidth,1);
      double ttmp = IntersectLine (ray,
				   fX - fLength *Math.cos (0.5*Math.PI+fDirection)+
				        fWidth*Math.cos (fDirection),
				   fY + fLength*Math.sin (0.5*Math.PI+fDirection)-
				        fWidth*Math.sin (fDirection),
				   (- Math.cos (fDirection)),
				   (+ Math.sin (fDirection)),
				 fWidth,1);
      t1 = Math.min(t1,ttmp);
      double t2 = Double.MAX_VALUE;

      for (int i = 0; i < 2; i++)      
	{
	    ttmp = IntersectCircle (ray,
				    fCircX[i],fCircY[i],
				    fCircRadius[i],
				    fBeginAngle[i],fEndAngle[i],2);
	    t2 = Math.min(t2,ttmp);
	}
      
      return Math.min (t1,t2);
    }
  
  public HitObject GetHit (LightRay ray)
    {
      double t1a = IntersectLine (ray,
				 fX + fLength *Math.cos (0.5*Math.PI+fDirection)+
				      fWidth*Math.cos (fDirection),
				 fY - fLength*Math.sin (0.5*Math.PI+fDirection)-
				      fWidth*Math.sin (fDirection),
				 (- Math.cos (fDirection)),
				 (+ Math.sin (fDirection)),
				 fWidth,1);
      
      double t1b = IntersectLine (ray,
				 fX - fLength *Math.cos (0.5*Math.PI+fDirection)+
				      fWidth*Math.cos (fDirection),
				 fY + fLength*Math.sin (0.5*Math.PI+fDirection)-
				      fWidth*Math.sin (fDirection),
				 (- Math.cos (fDirection)),
				 (+ Math.sin (fDirection)),
				 fWidth,1);
      
      double t2 = IntersectCircle (ray,
				   fCircX[0],fCircY[0],
				   fCircRadius[0],
				   fBeginAngle[0],fEndAngle[0],1);

      double t3 = IntersectCircle (ray,
				   fCircX[1],fCircY[1],
				   fCircRadius[1],
				   fBeginAngle[1],fEndAngle[1],1);

      if (t1a < t1b && t1a < t2 && t1a < t3)
	return new HitGlass (IntersectHitLine (ray,
					       fX + fLength*Math.cos (0.5*Math.PI+fDirection),
					       fY - fLength*Math.sin (0.5*Math.PI+fDirection),
					       (- Math.cos (fDirection)),
					       (+ Math.sin (fDirection)),
					       1,Math.PI+Math.PI*.5-fDirection),
			     GetRefractionConsts ());
      else if (t1b < t2 && t1b < t3)
	return new HitGlass (IntersectHitLine (ray,
					       fX - fLength*Math.cos (0.5*Math.PI+fDirection),
					       fY + fLength*Math.sin (0.5*Math.PI+fDirection),
					       (- Math.cos (fDirection)),
					       (+ Math.sin (fDirection)),
					       1,Math.PI*.5-fDirection),
			     GetRefractionConsts ());
      else if (t2 < t3)
	return new HitGlass (IntersectHitCircle (ray,
						 fCircX[0],fCircY[0],
						 fCircRadius[0],
						 fBeginAngle[0],fEndAngle[0],
						 2,true),
			     GetRefractionConsts ());
      else
	return new HitGlass (IntersectHitCircle (ray,
						 fCircX[1],fCircY[1],
						 fCircRadius[1],
						 fBeginAngle[1],fEndAngle[1],
						 2,true),
			     GetRefractionConsts ());
    }
}
