import java.awt.*;
import java.util.*;
import java.lang.*;
import OpticsObject;
import OpticsObjPanel;
import GlassPanel;

/*

n = sqrt (1 + a / (b - lambda^-2))

Crown glass:              a = 0.00012959897448336221437  b = 9.7515296756364723981e-5
Dense flint glass:        a = 7.54092574256e-5           b = 3.92374098659e-5
Light flint glass:        a = 7.1460083915e-5            b = 4.659619976479e-5
Crystal quartz:           a = 0.0001037                  b = 7.8280
Borosilicate crown glass: a = 9.8626e-5                  b = 7.8226e-5
Acrylic plastic:          a = 6.35411e-5                 b = 5.571192e-5
Vitreous quartz:          a = 8.297581018195e-5          b = 7.695266777784717e-5

 */


abstract class GlassObject extends OpticsObject
{
  double fRefractionIndex[] = { 1.6, 1.5 };
  double fRefractionConsts[] = new double[2];
  int nMaterial = 0;  
  static final double fMatrRefrConsts[][] = 
  { {0.00012959897448336221437 ,9.7515296756364723981e-5 },
    {7.54092574256e-5          ,3.92374098659e-5 },
    {7.1460083915e-5           ,4.659619976479e-5 },
    {0.0001037                 ,7.8280 },
    {9.8626e-5                 ,7.8226e-5 },
    {6.35411e-5                ,5.571192e-5 },
    {8.297581018195e-5         ,7.695266777784717e-5 } };
  

  public double[] GetRefractionConsts ()
    {
      if (nMaterial != -1)
	return fMatrRefrConsts[nMaterial];
      return fRefractionConsts;
    }

  public void SetPanel (OpticsObjPanel panel)
    {
      super.SetPanel (panel);
      if (thePanel != null)
	{
	  ((GlassPanel) thePanel).SetRefractionIndex (fRefractionIndex);
	  ((GlassPanel) thePanel).SetMaterial (nMaterial);
	}
    }

  public void CalcRefractionConsts ()
    {
      double nfs = fRefractionIndex[0]*fRefractionIndex[0];
      double nss = fRefractionIndex[1]*fRefractionIndex[1];

      fRefractionConsts[0] = (nfs-1)*(nss-1)/(288000*(nfs-nss));
      fRefractionConsts[1] = (-5+9*nfs-4*nss)/(1440000*(nfs-nss));
    }

  public void SetRefractionIndex (double index[],boolean setPanel)
    {
      theLabCanvas.AbortCalculation ();
      fRefractionIndex[0] = index[0];
      fRefractionIndex[1] = index[1];
      // System.out.println (""+fRefractionIndex[0]+"   "+fRefractionIndex[1]);
      CalcRefractionConsts ();
      if (setPanel && thePanel != null)
	((GlassPanel) thePanel).SetRefractionIndex (fRefractionIndex);
      theLabCanvas.repaint ();
      theLabCanvas.Recalculate ();
    }
  
  public void SetMaterial (int matr,boolean setPanel)
    {
      theLabCanvas.AbortCalculation ();
      nMaterial = matr;  
      if (setPanel && thePanel != null)
	((GlassPanel) thePanel).SetMaterial (nMaterial);
      theLabCanvas.repaint ();
      theLabCanvas.Recalculate ();
    }
}
