import java.awt.*;
import java.util.*;
import java.awt.event.*;
import OpticsObjPanel;

import LabCanvasPopupMenu;
import RayCalculator;
import LightRay;

class LabCanvas extends Canvas implements MouseListener, MouseMotionListener, ActionListener, ItemListener
{
  Vector theObjects = new Vector ();
  Panel theSettings;
  CardLayout theSettingsCardLayout;
  Hashtable thePanelsForObject;
  java.awt.List theObjectsList;
  Vector objectClasses;
  Point lastPopupMenuPoint;
  boolean bAutoRecalc = true;
  boolean bCleared = false;  

  LabCanvasPopupMenu thePopupMenu;

  int nActiveObject;
  Vector handlePoints;
  int nActiveHandlePoint;

  RayCalculator theCalculator = null;
  LightRay theRays[];

  LabCanvas (Vector classes,
	     Panel settings,
	     CardLayout settingsCardLayout,
	     Hashtable panelForObject,
	     java.awt.List list)
  {
    setBackground (Color.black);

    objectClasses = classes;

    addMouseListener (this);
    addMouseMotionListener (this);
    thePopupMenu = new LabCanvasPopupMenu (this,objectClasses);
    
    add(thePopupMenu);

    theSettings = settings;
    theSettingsCardLayout = settingsCardLayout;
    thePanelsForObject = panelForObject;
    theObjectsList = list;
    theObjectsList.addItemListener (this);

    ActivateObject (-1,true);
  }

  public void mousePressed(MouseEvent e)
  {
    nActiveHandlePoint = -1;
    if ((e.getModifiers () & InputEvent.BUTTON3_MASK) == 
	InputEvent.BUTTON3_MASK)
      {	
	lastPopupMenuPoint = new Point (e.getX(),e.getY());
	thePopupMenu.show (this,e.getX(),e.getY());
      }
    if ((e.getModifiers () & InputEvent.BUTTON2_MASK) != 
	InputEvent.BUTTON2_MASK &&
	(e.getModifiers () & InputEvent.BUTTON3_MASK) != 
	InputEvent.BUTTON3_MASK)
      {
	OpticsObject obj = GetActiveOpticsObject ();

	if (obj != null)
	  {
	    if (!obj.IsInsideCircle (e.getPoint ()))
	      {
		if (!ActivateObject (e.getPoint ()))
		  ActivateObject (-1,true);
		return;    
	      }
	    // handle selection of handle

	    int x = e.getX()-obj.nLabX;
	    int y = e.getY()-obj.nLabY;

	    if (handlePoints != null)
	      for (int i = handlePoints.size ()-1; i >= 0; i--)
		{
		  Point pt = (Point) handlePoints.elementAt (i); 

		  if ((pt.x-x)*(pt.x-x)+(pt.y-y)*(pt.y-y) <= 22)
		    {
		      nActiveHandlePoint = i;
		      obj.HandlePointMove (nActiveHandlePoint,e.getPoint ());
		    }
		}
	  }
	else // find an object to activate
	  ActivateObject (e.getPoint ());	  
      }    
  }

  public void mouseDragged(MouseEvent e)
  {
    if (nActiveHandlePoint == -1)
      return;
    OpticsObject obj = GetActiveOpticsObject ();
    obj.HandlePointMove (nActiveHandlePoint,e.getPoint ());
    handlePoints = obj.GetHandlePoints ();
  }
  
  public void mouseReleased(MouseEvent e)
  {
    if ((e.getModifiers () & InputEvent.BUTTON2_MASK) == 
	InputEvent.BUTTON2_MASK ||
	(e.getModifiers () & InputEvent.BUTTON3_MASK) == 
	InputEvent.BUTTON3_MASK ||
	nActiveHandlePoint == -1)
      return;
    OpticsObject obj = GetActiveOpticsObject ();
    obj.HandlePointMove (nActiveHandlePoint,e.getPoint ());
    handlePoints = obj.GetHandlePoints ();
    nActiveHandlePoint = -1;
  }

  public void mouseMoved(MouseEvent e) { }
  public void mouseEntered(MouseEvent e) { }
  public void mouseExited(MouseEvent e) { }
  public void mouseClicked(MouseEvent e) { }

  boolean ActivateObject (Point pt)
  {
    double fBestLength = Double.MAX_VALUE;
    int nBestComponent = -1;

    for (int i = 0; i < theObjects.size (); i++)
      if (GetOpticsObject (i).IsInsideCircle (pt))
	{
	  double length = GetOpticsObject (i).GetPointLength (pt);
	  
	  if (length < fBestLength)
	    {
	      fBestLength = length;
	      nBestComponent = i;
	    }
	}
    if (nBestComponent != -1)
      {
	ActivateObject (nBestComponent,true);
	return true;
      }
    return false;
  }

  public void actionPerformed(ActionEvent e)
  {
    Object source = e.getSource ();

    if (source == thePopupMenu.miRemove)
      {
	OpticsObject obj = GetActiveOpticsObject ();
	
	if (obj != null)
	  {
	    AbortCalculation ();
	    theObjects.removeElementAt (nActiveObject);
	    theObjectsList.remove (nActiveObject);    
	    Recalculate ();
	  }
	ActivateObject (-1,true);
      }
    int index = thePopupMenu.miAdds.indexOf (source);
    if (index != -1)
      {
	try
	  {
	    OpticsObject obj = 
	      (OpticsObject) ((Class) objectClasses.
			      elementAt (index)).newInstance ();

	    obj.SetLabCanvas (this);
	    
	    obj.SetLocation (lastPopupMenuPoint.x/10.0,
			     lastPopupMenuPoint.y/10.0,false);
	    
	    String name = 
	      (String) obj.getClass ().getMethod ("GetUniqueName",
						  null).invoke (obj,null);
	    
	    theObjects.addElement (obj);
	    theObjectsList.add (name);

	    ActivateObject (theObjects.size ()-1,true);
	    Recalculate ();
	  }
	catch ( Exception ex )
	  {
	    System.out.println ("Unable to create: "+
				((Class) objectClasses.
				 elementAt (index)).
				getName ());
	    ex.printStackTrace ();
	  }
      }
  }

  public void itemStateChanged(ItemEvent e)
  {
    Object source = e.getSource ();

    if (source == theObjectsList)
      ActivateObject (theObjectsList.getSelectedIndex (),false);
  }

  OpticsObject GetOpticsObject (int nIndex)
  {
    if (nIndex == -1)
      return null;
    return (OpticsObject) theObjects.elementAt (nIndex);
  }

  OpticsObject GetActiveOpticsObject ()
    {
      return GetOpticsObject (nActiveObject);
    }

  void ActivateObject (int nObject,boolean setListItem)
  {
    OpticsObject obj = GetOpticsObject (nObject);
    if (obj != null)
      obj.SetPanel (null);
    nActiveObject = nObject;

    if (setListItem)
      theObjectsList.select (nActiveObject);
    obj = GetActiveOpticsObject ();
    if (obj == null)
      {
	theSettingsCardLayout.show (theSettings,"Nothing");
	handlePoints = null;
      }
    else
      {
	try
	  {
	    OpticsObjPanel panel = 
	      (OpticsObjPanel) thePanelsForObject.get (obj.getClass ().
						       getName ());
	    theSettingsCardLayout.show (theSettings,
					(String) panel.getClass ().getName ());
	    obj.SetPanel (panel);
	  }
	catch ( Exception ex )
	  {
	    ex.printStackTrace ();
	  }
	handlePoints = obj.GetHandlePoints ();
      }
    repaint ();
  }

  public void paint (Graphics g)
  {
    int w = getSize ().width;
    int h = getSize ().height;
    
    if (theRays != null)
      {
	for (int i = 0; i < theRays.length; i++)
	  {
	    LightRay ray = theRays[i]; // (LightRay) theRays.elementAt (i);

	    g.setColor (ray.GetColor ());

	    double x1 = ray.pX*10.0;
	    double y1 = ray.pY*10.0;
	    double x2 = ray.vX*10.0;
	    double y2 = ray.vY*10.0;
	    
	    if (x1 < 0) { y1 += (y2-y1)*( -x1)/(x2-x1); x1 = 0; }
	    if (x2 < 0) { y2 += (y1-y2)*( -x2)/(x1-x2); x2 = 0; }
	    if (x1 > w) { y1 += (y2-y1)*(w-x1)/(x2-x1); x1 = w; }
	    if (x2 > w) { y2 += (y1-y2)*(w-x2)/(x1-x2); x2 = w; }
	    if (y1 < 0) { x1 += (x2-x1)*( -y1)/(y2-y1); y1 = 0; }
	    if (y2 < 0) { x2 += (x1-x2)*( -y2)/(y1-y2); y2 = 0; }
	    if (y1 > h) { x1 += (x2-x1)*(h-y1)/(y2-y1); y1 = h; }
	    if (y2 > h) { x2 += (x1-x2)*(h-y2)/(y1-y2); y2 = h; }
	    
	    g.drawLine ((int) x1,(int) y1,(int) x2,(int) y2);
	  }
      }

    // draw the object contours
    for (int i = 0; i < theObjects.size (); i++)
      GetOpticsObject (i).DrawContours (g,i == nActiveObject);

    OpticsObject obj = GetActiveOpticsObject ();
    if (obj != null)
      {
	  Graphics g2 = g.create (obj.nLabX-obj.nLabRadius-4,obj.nLabY-obj.nLabRadius-4,
				  obj.nLabRadius*2+8,obj.nLabRadius*2+8);
	  g2.translate (obj.nLabRadius+4,obj.nLabRadius+4);
	
	  g2.setColor (Color.red);
	  
	  if (handlePoints != null)
	      for (int i = 0; i < handlePoints.size (); i++)
		  {
		      Point pt = (Point) handlePoints.elementAt (i); 
		      
		      if (i == 0)
			  {
			      g2.drawOval (pt.x-4,pt.y-4,8,8);
			      g2.drawLine (pt.x-2,pt.y-2,pt.x+2,pt.y+2);
			      g2.drawLine (pt.x+2,pt.y-2,pt.x-2,pt.y+2);
			  }
		      else
			  {
			      g2.drawRect (pt.x-3,pt.y-3,6,6);
			      g2.drawLine (pt.x-3,pt.y-3,pt.x+3,pt.y+3);
			      g2.drawLine (pt.x+3,pt.y-3,pt.x-3,pt.y+3);
			  }
		  }
      }
    
    g.setColor (Color.lightGray);
    g.draw3DRect (0,0,w-1,h-1,false);
  }
    
  public void AbortCalculation ()
  {
    theRays = null;
    if (theCalculator != null)
      {
	if (theCalculator.isAlive ())
	  theCalculator.stop ();
	theCalculator = null;
      }
  }

  public void Recalculate () 
  {
    Recalculate(false);
  }

  public void Recalculate (boolean bForceCalc) 
  {
    // kill any calculation going on and start all over again

      AbortCalculation ();

    if (bAutoRecalc || bForceCalc)
      {
	theCalculator = new RayCalculator (this,theObjects);
	theCalculator.setDaemon (true);
	if (theCalculator.getPriority () > Thread.MIN_PRIORITY)
	  theCalculator.setPriority (theCalculator.getPriority ()-1);
	theCalculator.start ();
	bCleared = false;
      }
    else if (!bCleared)
      {
	bCleared = true;
	for (int i = 0; i < theObjects.size (); i++)
	  {
	    OpticsObject obj = (OpticsObject) theObjects.elementAt (i);
	    
	    obj.NewTrace ();
	    obj.TraceFinished ();
	  }
      }
  }

  public void CalculationFinished (LightRay rays[])
  {
    theRays = rays;
    repaint ();

    RayCalculator killCalc = theCalculator;
    theCalculator = null;
    if (killCalc.isAlive ())
	killCalc.stop ();
  }
}

