import java.lang.*;
import java.awt.Color;

class LightRay
{
  public static final float MIN_WAVE_LENGTH = 400;
  public static final float MAX_WAVE_LENGTH = 750;

  public double pX, pY;
  public double vX, vY;

  public int nBounces;

  public Object lastHitObject;
  public int lastHitSurface;

  public float fEfieldParallell;
  public float fEfieldPerpendicular;

  public float fIntensity;

  public float fWaveLength; /* MIN_WAVE_LENGTH-MAX_WAVE_LENGTH
				negative (integer) is no of 'white' rays
				*/
  
  LightRay (double x,double y,double vx,double vy,float waveLength)
  {
    pX = x;
    pY = y;
    vX = vx;
    vY = vy;

    lastHitObject = null;

    fEfieldParallell = 1.0f;
    fEfieldPerpendicular = 1.0f;

    fIntensity = 1.0f;

    fWaveLength = waveLength;
  }

  LightRay (double x,double y,double vx,double vy,
	    int bounces,Object lastObj,int lastSurface,
	    float eFParallell,float eFPerpendicular,
	    float waveLength)
  {
    pX = x;
    pY = y;
    vX = vx;
    vY = vy;
    nBounces = bounces;
    lastHitObject = lastObj;
    lastHitSurface = lastSurface;
    fEfieldParallell = eFParallell;
    fEfieldPerpendicular = eFPerpendicular;

    fIntensity = (float) ((fEfieldParallell*
			   fEfieldParallell +
			   fEfieldPerpendicular*
			   fEfieldPerpendicular)/2.0);
      
      if (fIntensity > 1.0)
	fIntensity = (float) 1.0;

    fWaveLength = waveLength;
  }

    float InterP (float vals[][],float fFrac)
    {
	int i = 1;
	
	for (; vals[i][0] < fFrac; i++);
	
	return vals[i-1][1]+
	    ((fFrac-vals[i-1][0])*
	     (vals[i][1]-vals[i-1][1])/
	     (vals[i][0]-vals[i-1][0]));
    }
    
    static final float redvals[][] = {
	{0,255},{345,255},{400,235},{435,215},{475,155},{500,105},
	{515,30},{530,0},{985,0},{990,30},{1015,90},{1050,130},{1120,185}};
    static final float greenvals[][] = {
	{0,0},{25,0},{35,60},{80,120},{130,160},{180,190},{230,215},{315,255},
	{685,255},{750,230},{820,190},{880,155},{930,100},{985,25},
	{990,0},{1185,0},{1190,25},{1100,0},{1120,0}};
    static final float bluevals[][] = {
	{0,0},{350,0},{410,25},{450,25},{500,0},{515,0},{520,70},{560,170},{600,230},
	{685,255},{990,255},{1045,230},{1120,195}};
    
    static final float wavevals[][] = { 
	{400,1120}, 
	{450,1050},
	{490,700},
	{575,430},
	{595,250},
	{620,50},
	{750,0} };
    
/* violett     400-450
   blue        450-500
   green       500-570
   yellow      570-590
   orange      590-610
   red         610-750
*/
  Color GetColor ()
    {     
      if (fWaveLength <= 0.0)
	return new Color(fIntensity,fIntensity,fIntensity);
      else
	{
	  float fFrac = fWaveLength > MIN_WAVE_LENGTH ? 
	      (fWaveLength < MAX_WAVE_LENGTH ? (fWaveLength) : MAX_WAVE_LENGTH) : MIN_WAVE_LENGTH;

	  fFrac = InterP (wavevals,fFrac);
	  
	  return new Color ((float) (InterP(redvals,fFrac)*fIntensity/255.0),
			    (float) (InterP(greenvals,fFrac)*fIntensity/255.0),
			    (float) (InterP(bluevals,fFrac)*fIntensity/255.0));
	  
	      
	  /*
	    return new Color((float) ((MAX_WAVE_LENGTH-fWaveLength)/
	    (MAX_WAVE_LENGTH-MIN_WAVE_LENGTH)*fIntensity),
	    (float) ((fWaveLength-MIN_WAVE_LENGTH)/
	    (MAX_WAVE_LENGTH-MIN_WAVE_LENGTH)*fIntensity),
	    fIntensity);
	  */
	}
    }
}
