import java.applet.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;

import LabCanvas;

import PointLightSourcePanel;
import PointLightSourceObject;
import ParallellLightSourcePanel;
import ParallellLightSourceObject;
import PaperPanel;
import PaperObject;
import PlaneMirrorPanel;
import PlaneMirrorObject;
import CircularMirrorPanel;
import CircularMirrorObject;
import EquilateralTrianglePrismPanel;
import EquilateralTrianglePrismObject;
import CircularLensPanel;
import CircularLensObject;
import GeneralLensPanel;
import GeneralLensObject;

public class Optics extends java.applet.Applet implements Runnable, ActionListener, ItemListener
{
  Thread kicker;

  LabCanvas lab;

  Button buttonRecalc;
  Checkbox checkboxAutoRecalc;
  
  java.awt.List listObjects;

  Panel settings;
  CardLayout settingsCardLayout;
  Hashtable panelForObject = new Hashtable ();

  Vector objectClasses = new Vector ();

  public void init()
    {
      try
	{
	  CreateUI ();
	}
      catch ( Throwable ex )
	{
	  setLayout (new GridLayout (1,1));
	  TextArea a = new TextArea ();
	  add(a);
	  a.setText ("Init failed.\n\nAn unknown error has occurred.\n\nI've experienced some error when\nrunning the program on systems that cannot\nhandle java 1.1.5 (or something like that (I know\nthat intepreters capable of 1.1.2 fail)).\nThis is probably the cause of this problem.\n\nSolutions:\n\n* Get a capable browser/appletviewer. This program is available as\nhttp://www.dd.chalmers.se/~f96hajo/Optics/optics.zip\n(use the file optics.html from the zip-file to start)\n\n* If you are located at dd.chalmers.se, you may run\nappletviewer ~f96hajo/WWW/optics/optics.html\nfrom an Xterm-window.\n\n");
	  
	  if (ex instanceof NoSuchMethodException )
	    System.out.println ("init::Error");
	  else
	    System.out.println ("init::General init error");
	  
	}
  }
  
  protected void CreateUI ()
    {
      //		setSize(200,150);
      
      setBackground (Color.lightGray);
      
      setLayout(new BorderLayout());
	
      Panel westPanel = new Panel ();
      westPanel.setLayout (new BorderLayout());
      
      Panel ctrlPanel = new Panel ();
      Panel ctrlPanelButtons = new Panel ();
      ctrlPanel.setLayout (new BorderLayout ());
      ctrlPanelButtons.setLayout (new GridLayout (0,1));
      
      westPanel.add("North",ctrlPanel);
      
      settings = new Panel ();
      settings.setLayout (settingsCardLayout = new CardLayout ());
      westPanel.add ("Center",settings);
      
      ctrlPanel.add ("North",ctrlPanelButtons);
      ctrlPanelButtons.add (buttonRecalc = new Button ("Recalc"));
      ctrlPanelButtons.add (checkboxAutoRecalc = new Checkbox ("Auto-recalc"));
      ctrlPanel.add ("South",listObjects = new java.awt.List ());
      
      checkboxAutoRecalc.setState (true);
      buttonRecalc.addActionListener (this);
      checkboxAutoRecalc.addItemListener (this);

      settings.add (new Panel (),"Nothing");	
	
      AddObjectClass ("PointLightSourceObject");
      AddObjectClass ("ParallellLightSourceObject");
      AddObjectClass ("PaperObject");
      AddObjectClass ("PlaneMirrorObject");
      AddObjectClass ("CircularMirrorObject");
      AddObjectClass ("EquilateralTrianglePrismObject");
      AddObjectClass ("CircularLensObject");
      AddObjectClass ("GeneralLensObject");
      
      add("Center",lab = new LabCanvas(objectClasses,
				       settings,
				       settingsCardLayout,
				       panelForObject,
				       listObjects));
      add("West",westPanel);
    }
  
  void AddObjectClass (String name)
    {
      try
	{
	  Class c = Class.forName (name);
	  
	  objectClasses.addElement (c);
	  
	  String panelName = (String) c.getMethod ("GetPanelName",
						   null).invoke (null,null);
	  
	  Class cPanel = Class.forName (panelName);
	  
	  OpticsObjPanel panel = (OpticsObjPanel) cPanel.newInstance ();
	  
	  panelForObject.put (c.getName (),panel);
	  
	  settings.add (panel,panelName);		
	}
      catch ( Exception ex)
	{
	  System.out.println ("Failed to load: "+name);
	  ex.printStackTrace ();
      }
      
    }
  
  public void start()
    {
      if (kicker == null)
	{
	kicker = new Thread(this);
	kicker.start ();
	}
    }
  
  public void stop()
    {
      kicker.stop ();
      kicker = null;
    }                     
  
  public void run()
    {
      for ( ; ; )
	{
	  try
	    {
	    Thread.sleep (1000);			
	    }
	  catch (InterruptedException e)
	    {
	      
	    }    	                    	
	}
    }
  
  public void actionPerformed(ActionEvent e)
    {
      Object source = e.getSource ();
      
      if (source == buttonRecalc)
      {
	lab.Recalculate (true);
      }
    }
  
  public void itemStateChanged(ItemEvent e)
    {
      Object source = e.getSource ();
      
      if (source == checkboxAutoRecalc)
      {
	lab.bAutoRecalc = checkboxAutoRecalc.getState ();
	lab.Recalculate (true);
      }
    }
  
  public String getAppletInfo ()
    {
      return "Optics Applet by Hkan T. Johansson, Sweden, 1998.";
    }
}

