import java.awt.*;
import java.util.*;
import OpticsObjPanel;
import LabCanvas;
import LightRay;
import HitObject;
import HitObjectNormal;

abstract class OpticsObject
{
  OpticsObjPanel thePanel = null;
  LabCanvas theLabCanvas = null;

  double fX = 0.0;
  double fY = 0.0;
  double fDirection = 0.0;
  
  public double fRadius; // need only be maintained by non-sources
  public int nLabX, nLabY, nLabRadius;

  OpticsObject()
  {

  }

  public void SetLabCanvas (LabCanvas canvas)
  {
    theLabCanvas = canvas;
  }

  public void SetPanel (OpticsObjPanel panel)
  {
    if (panel == null && thePanel != null)
      thePanel.theObject = null;
    thePanel = panel;
    if (thePanel != null)
      {
	thePanel.theObject = this;

	thePanel.SetLocation (fX,fY);
	thePanel.SetDirection (fDirection);



      }
  }

  public void SetLocation (double x,double y,boolean setPanel)
  {
    theLabCanvas.AbortCalculation ();
    fX = x;
    fY = y;

    nLabX = (int) (fX * 10.0);
    nLabY = (int) (fY * 10.0);

    if (setPanel && thePanel != null)
      thePanel.SetLocation (fX,fY);
    UpdateInternalState ();
    theLabCanvas.repaint ();
    theLabCanvas.Recalculate ();
  }

  public void SetDirection (double direction,boolean setPanel)
  {
    theLabCanvas.AbortCalculation ();
    fDirection = direction;
    if (setPanel && thePanel != null)
      thePanel.SetDirection (fDirection);
    UpdateInternalState ();
    theLabCanvas.repaint ();
    theLabCanvas.Recalculate ();
  }

  protected void UpdateInternalState() { }
  
  int RadiansToIntDegrees (double radians)
  {
    return (int) (radians*180.0/Math.PI);
  }

  public Graphics InitDrawContours (Graphics g,boolean selected)
  {
    g = g.create (nLabX-nLabRadius-4,nLabY-nLabRadius-4,
		  nLabRadius*2+8,nLabRadius*2+8);
    g.translate (nLabRadius+4,nLabRadius+4);
    g.setColor(selected ? Color.orange : Color.cyan);

    return g;
  }

  public boolean IsInsideCircle (Point pt)
  {
    return (pt.x-nLabX)*(pt.x-nLabX)+(pt.y-nLabY)*(pt.y-nLabY) <= 
      (nLabRadius+4)*(nLabRadius+4);
  }

  public Vector GetHandlePoints ()
  {
    Vector v = new Vector ();
    v.addElement (new Point (0,0));
    return v;
  }

  public double GetPointDirection (Point pt)
  {
    return Math.atan2(-(pt.y-nLabY),pt.x-nLabX);
  }

  public double GetPointLength (Point pt)
  {
    int x = pt.x-nLabX;
    int y = pt.y-nLabY;

    return Math.sqrt(x*x+y*y)/10.0;
  }

  public double GetPointLength (Point pt,double dx,double dy)
  {
    int x = pt.x-nLabX;
    int y = pt.y-nLabY;

    return Math.sqrt((x-dx*10)*(x-dx*10)+(y-dy*10)*(y-dy*10))/10.0;
  }

  public double GetPointLengthDir (Point pt,double vx,double vy)
  {
    int x = pt.x-nLabX;
    int y = pt.y-nLabY;

    return (x*vx+y*vy)/10.0;
  }

  public double GetPointAngle (Point pt)
  {
    return ToPiInterval(GetPointDirection(pt)-fDirection);
  }

  public double GetPointX (Point pt)
  {
    return (pt.x-nLabX)/10.0;
  }

  public double GetPointY (Point pt)
  {
    return (pt.y-nLabY)/10.0;
  }

  double ToPiInterval (double rad)
  {
    while (rad > Math.PI)
      rad -= 2*Math.PI;
    while (rad < -Math.PI)
      rad += 2*Math.PI;
    return rad;
  }

  double ToAboveZeroInterval (double rad)
  {
    while (rad > 2*Math.PI)
      rad -= 2*Math.PI;
    while (rad < 0.0)
      rad += 2*Math.PI;
    return rad;
  }

  public void HandlePointMove (int nPoint,Point pt)
  {
    if (nPoint == 0)
      {
	SetLocation (pt.x/10.0,pt.y/10.0,true);
      }    
  }

  public void GetRays(Vector raysToTrace) {}
  public void NewTrace() {}
  public void TraceFinished() {}

  public boolean IsSource ()
  {
    return false;
  }

  public double IsHit (LightRay ray)
  {
    return 0.0;
  }

  public HitObject GetHit (LightRay ray)
  {
    return null;
  }

  protected double IntersectLinePart (LightRay ray,
				      double x1,double y1,
				      double vx,double vy,
				      double fLength)
  {
      double t = (vx*(ray.pY-y1)-vy*(ray.pX-x1))/(ray.vX*vy-ray.vY*vx);
      double s = (Math.abs(vx) > Math.abs(vy) ? 
		  (ray.pX+t*ray.vX-x1)/vx : 
		  (ray.pY+t*ray.vY-y1)/vy);
      return s / (2*fLength);
  }

  protected double IntersectLine (LightRay ray,
				  double x1,double y1,
				  double vx,double vy,
				  double fLength,
				  int surface)
    {
      if (ray.lastHitObject == this && ray.lastHitSurface == surface)
	return Double.MAX_VALUE;
      
      double t = (vx*(ray.pY-y1)-vy*(ray.pX-x1))/(ray.vX*vy-ray.vY*vx);
      if (t < 0.0) // intersection behind ray-start
	  return Double.MAX_VALUE;
      double s = (Math.abs(vx) > Math.abs(vy) ? 
		  (ray.pX+t*ray.vX-x1)/vx : 
		  (ray.pY+t*ray.vY-y1)/vy);
      if (s < 0.0 || s > 2*fLength) // intersection outside
	  return Double.MAX_VALUE;
      return t;
    }
  
  protected HitObjectNormal IntersectHitLine (LightRay ray,
					      double x1,double y1,
					      double vx,double vy,
					      int surface,
					      double normal)
    {
      double t = (vx*(ray.pY-y1)-vy*(ray.pX-x1))/(ray.vX*vy-ray.vY*vx);
      /*      double s = (Math.abs(vx) > Math.abs(vy) ? 
		  (ray.pX+t*ray.vX-x1)/vx : 
		  (ray.pY+t*ray.vY-y1)/vy);
		  */
      return new HitObjectNormal (ray.pX+t*ray.vX,ray.pY+t*ray.vY,surface,normal);
    }

  public boolean IsValidCircleIntersection (double x,double y,
					    double mx,double my,
					    double begin,double end)
    {
      double fAngle = Math.atan2 (-(y-my),x-mx);
      
      if (fAngle < begin)
	fAngle += 2*Math.PI;
      
      return fAngle < end;
    }
  
  protected double IntersectCircle (LightRay ray,
				    double x,double y,
				    double radius,
				    double begin,double end,
				    int surface)
    {
      double p = 2*(ray.vX*(ray.pX-x)+ray.vY*(ray.pY-y));
      double q = (ray.pX-x)*(ray.pX-x)+
	(ray.pY-y)*(ray.pY-y)-
	radius*radius;
      
      double sqr = p*p/4-q;
      
      if (sqr < 0.0) // no intersection
	return Double.MAX_VALUE;
      
      sqr = Math.sqrt(sqr);
      p /= -2;
      
      double t = Double.MAX_VALUE;
      double t1 = p+sqr;
      if (t1 < 0.0) // both intersections behind ray
	return Double.MAX_VALUE;
      
      if (p-sqr > 0.0)
	{
	  double t2 = p-sqr;
	  
	  if (t2 > 1.0e-8)
	    {
	      double hX = ray.pX+t2*ray.vX;
	      double hY = ray.pY+t2*ray.vY;
	      
	      if (IsValidCircleIntersection (hX,hY,x,y,begin,end))
		t = t2;
	    }
	}
      if (t == Double.MAX_VALUE && t1 > 1.0e-8)
	{
	  double hX = ray.pX+t1*ray.vX;
	  double hY = ray.pY+t1*ray.vY;
	  
	  if (IsValidCircleIntersection (hX,hY,x,y,begin,end))
	    t = t1;
	}
      
      if (t < 1.0e-8) // we've hit ourselves, avoid infinite trouble
	return Double.MAX_VALUE; 
      
      return t;
    }

  protected HitObjectNormal IntersectHitCircle (LightRay ray,
						double x,double y,
						double radius,
						double begin,double end,
						int surface,boolean bInvert)
    {
      double p = 2*(ray.vX*(ray.pX-x)+ray.vY*(ray.pY-y));
      double q = (ray.pX-x)*(ray.pX-x)+
	(ray.pY-y)*(ray.pY-y)-
	radius*radius;
      
      double sqr = p*p/4-q;
      
      sqr = Math.sqrt(sqr);
      p /= -2;
      
      double t = 0.0;
      
      if (p-sqr > 0.0)
	{
	  double t2 = p-sqr;
	  
	  if (t2 > 1.0e-8)
	    {
	      double hX = ray.pX+t2*ray.vX;
	      double hY = ray.pY+t2*ray.vY;
	      
	      if (IsValidCircleIntersection (hX,hY,x,y,begin,end))
		t = t2;
	    }
	}
      double t1 = p+sqr;
      if (t == 0.0 && t1 > 1.0e-8)
	{
	  double hX = ray.pX+t1*ray.vX;
	  double hY = ray.pY+t1*ray.vY;
	  
	  if (IsValidCircleIntersection (hX,hY,x,y,begin,end))
	    t = t1;
	}
      
      double hX = ray.pX+t*ray.vX;
      double hY = ray.pY+t*ray.vY;
      
      return new HitObjectNormal (hX,hY,surface,(bInvert ? Math.PI : 0) + (radius >= 0 ?
									   Math.PI-Math.atan2(-(hY-y),(hX-x)) :
									   -Math.atan2(-(hY-y),(hX-x))));
      
    }
    
  public abstract void DrawContours (Graphics g,boolean selected);
}

