import java.awt.*;
import java.util.*;
import java.lang.*;
import OpticsObject;
import OpticsObjPanel;
import LightRay;
import HitObject;
import HitPaper;

import java.lang.Math;

class PaperObject extends OpticsObject
{
  protected static final double MAX_LENGTH = 10.0;

  double fLength = 5.0;

  Vector vectorRaysCaught = null;

  static int nObjectNo = 1;
  public String GetUniqueName () { return "Paper "+(nObjectNo++); }
  public static String GetPanelName () { return "PaperPanel"; }
  public static String GetObjectName () { return "Paper"; }

  public void SetPanel (OpticsObjPanel panel)
  {
    super.SetPanel (panel);
    if (thePanel != null)
      {
	((PaperPanel) thePanel).SetLength (fLength);
	((PaperPanel) thePanel).SetObject (this);
      }
  }

  public void SetLength (double length,boolean setPanel)
  {
    theLabCanvas.AbortCalculation ();
    fLength = length;
    fRadius = fLength;
    nLabRadius = (int) (fLength * 10.0);
    if (setPanel && thePanel != null)
      ((PaperPanel) thePanel).SetLength (fLength);
    theLabCanvas.repaint ();
    theLabCanvas.Recalculate ();
  }

  public void DrawContours (Graphics g,boolean selected)
  {
    g = InitDrawContours (g,selected);

    double fXLength =  fLength*10.0*Math.cos (0.5*Math.PI+fDirection);
    double fYLength = -fLength*10.0*Math.sin (0.5*Math.PI+fDirection);

    double fXOffset =  6.0*Math.cos (fDirection);
    double fYOffset = -6.0*Math.sin (fDirection);

    g.drawLine ((int) fXLength,(int) fYLength,
		(int) -fXLength,(int) -fYLength);
    g.drawLine ((int) fXLength,(int) fYLength,
		(int) (fXOffset+fXLength),(int) (fYOffset+fYLength));
    g.drawLine ((int) -fXLength,(int) -fYLength,
		(int) (fXOffset-fXLength),(int) (fYOffset-fYLength));
  }

  public void HandlePointMove (int nPoint,Point pt)
  {
    super.HandlePointMove (nPoint,pt);

    if (nPoint == 1)
      {
	double fDirection = GetPointDirection (pt);
	double fLength = GetPointLength (pt);

	if (fLength > MAX_LENGTH)
	  fLength = MAX_LENGTH;
	SetDirection (ToPiInterval (fDirection-0.5*Math.PI),true);
	SetLength (fLength,true);
      }
  }
  
  public Vector GetHandlePoints ()
  {
    Vector v = super.GetHandlePoints ();
    
    v.addElement (new Point ((int) (( fLength*10.0*
				     Math.cos (0.5*Math.PI+fDirection))),
			     (int) ((-fLength*10.0*
				     Math.sin (0.5*Math.PI+fDirection)))));
    return v;
  }

  public void NewTrace()
  {
    vectorRaysCaught = null;
    vectorRaysCaught = new Vector ();
  }

  public void TraceFinished ()
  {
    if (thePanel != null)
      ((PaperPanel) thePanel).pdDisplay.repaint ();
  }

  public double IsHit (LightRay ray)
  {
    return IntersectLine (ray,
			  fX + fLength*Math.cos (0.5*Math.PI+fDirection),
			  fY - fLength*Math.sin (0.5*Math.PI+fDirection),
			  (- Math.cos (0.5*Math.PI+fDirection)),
			  (+ Math.sin (0.5*Math.PI+fDirection)),
			  fLength,1);
    /*
    double x1 = fX + fLength*Math.cos (0.5*Math.PI+fDirection);
    double y1 = fY - fLength*Math.sin (0.5*Math.PI+fDirection);
    double vx = - Math.cos (0.5*Math.PI+fDirection);
    double vy = + Math.sin (0.5*Math.PI+fDirection);

    double t = (vx*(ray.pY-y1)-vy*(ray.pX-x1))/(ray.vX*vy-ray.vY*vx);
    if (t < 0.0) // intersection behind ray-start
      return 0.0;
    double s = (Math.abs(vx) > Math.abs(vy) ? 
		(ray.pX+t*ray.vX-x1)/vx : 
		(ray.pY+t*ray.vY-y1)/vy);
    if (s < 0.0 || s > 2*fLength) // intersection outside paper
      return 0.0;
    return t;
    */
  }

  public HitObject GetHit (LightRay ray)
  {
      if (Math.cos(fDirection)*ray.vX-Math.sin(fDirection)*ray.vY > 0) // must hit front side
	  {
	      vectorRaysCaught.addElement (ray);
	      vectorRaysCaught.
		  addElement (new Double(IntersectLinePart(ray,
							   fX + fLength*Math.cos (0.5*Math.PI+fDirection),
							   fY - fLength*Math.sin (0.5*Math.PI+fDirection),
							   (- Math.cos (0.5*Math.PI+fDirection)),
							   (+ Math.sin (0.5*Math.PI+fDirection)),fLength)));
	  }
  
    return new HitPaper (IntersectHitLine (ray,
					   fX + fLength*
					   Math.cos (0.5*Math.PI+fDirection),
					   fY - fLength*
					   Math.sin (0.5*Math.PI+fDirection),
					   (- Math.cos (0.5*Math.PI+fDirection)),
					   (+ Math.sin (0.5*Math.PI+fDirection)),
					   1,Math.PI-fDirection));
    /*
    double x1 = fX + fLength*Math.cos (0.5*Math.PI+fDirection);
    double y1 = fY - fLength*Math.sin (0.5*Math.PI+fDirection);
    double vx = - Math.cos (0.5*Math.PI+fDirection);
    double vy = + Math.sin (0.5*Math.PI+fDirection);

    double t = (vx*(ray.pY-y1)-vy*(ray.pX-x1))/(ray.vX*vy-ray.vY*vx);
    double s = (Math.abs(vx) > Math.abs(vy) ? 
		(ray.pX+t*ray.vX-x1)/vx : 
		(ray.pY+t*ray.vY-y1)/vy);

    // add rays to Array for use by PaperDisplay

    return new HitPaper (ray.pX+t*ray.vX,ray.pY+t*ray.vY,0);
    */
  }
}
