import java.awt.*;
import java.lang.Math;
import java.util.*;

import LightSourceObject;
import ParallellLightSourcePanel;
import LightRay;

class ParallellLightSourceObject extends LightSourceObject
{
  protected static final double MAX_LENGTH = 10.0;

  double fLength = 1.0;

  ParallellLightSourceObject ()
  {
    nLabRadius = 20;
  }

  static int nObjectNo = 1;
  public String GetUniqueName () { return "Parallell Source "+(nObjectNo++); }
  public static String GetPanelName () { return "ParallellLightSourcePanel"; }
  public static String GetObjectName () { return "Parallell Light Source"; }

  public void SetPanel (OpticsObjPanel panel)
  {
    super.SetPanel (panel);
    if (thePanel != null)
      ((ParallellLightSourcePanel) thePanel).SetLength (fLength);
  }
  
  public void SetLength (double length,boolean setPanel)
  {
    theLabCanvas.AbortCalculation ();
    fLength = length;
    nLabRadius = (int) (fLength * 10.0)+10;
    if (setPanel && thePanel != null)
      ((ParallellLightSourcePanel) thePanel).SetLength (fLength);
    theLabCanvas.repaint ();
    theLabCanvas.Recalculate ();
  }

  public void DrawContours (Graphics g,boolean selected)
  {
    g = InitDrawContours (g,selected);

    double x = 10*-fLength*Math.sin(fDirection);
    double y = 10*-fLength*Math.cos(fDirection);

    double dx = 5*Math.cos(fDirection);
    double dy = -5*Math.sin(fDirection);

    g.drawLine ((int) x,(int) y,(int) -x,(int) -y);
    g.drawLine ((int) (x+dx),(int) (y+dy),(int) (-x+dx),(int) (-y+dy));
    g.drawLine ((int) (x-dx),(int) (y-dy),(int) (-x-dx),(int) (-y-dy));
    g.drawLine ((int) (x-2*dx),(int) (y-2*dy),(int) (-x-2*dx),(int) (-y-2*dy));

    g.drawArc ((int) x-10,(int) y-10,20,20,
	       RadiansToIntDegrees(fDirection),180);
    g.drawArc ((int) x-5,(int) y-5,10,10,
	       RadiansToIntDegrees(fDirection),180);
    g.drawArc ((int) -x-10,(int) -y-10,20,20,
	       180+RadiansToIntDegrees(fDirection),180);
    g.drawArc ((int) -x-5,(int) -y-5,10,10,
	       180+RadiansToIntDegrees(fDirection),180);

    /*    g.drawArc (-20,-20,
	       40,40,
	       RadiansToIntDegrees(fDirection+fWidth),
	       360-RadiansToIntDegrees(2*fWidth));
	       
    g.drawOval (-10,-10,
	       20,20);
    g.drawLine (-3,-3,3,3);
    g.drawLine (3,-3,-3,3);
    */
  }

  public void HandlePointMove (int nPoint,Point pt)
  {
    super.HandlePointMove (nPoint,pt);

    if (nPoint == 1)
      {
	double fDirection = GetPointDirection (pt);
	double fLength = GetPointLength (pt);

	if (fLength > MAX_LENGTH)
	  fLength = MAX_LENGTH;
	SetDirection (ToPiInterval (fDirection-0.5*Math.PI),true);
	SetLength (fLength,true);
      }
  }

  public void GetRays (Vector raysToTrace)
  { 
    for (int i = 0; i < nRays; i++)
      {
	double fDist = (nRays > 1 ? -fLength + fLength * 2 * i / (nRays-1) : 0);
	
	raysToTrace.addElement (new LightRay (fX+fDist*Math.sin(fDirection),
					      fY+fDist*Math.cos(fDirection),
					      Math.cos (fDirection),-Math.sin (fDirection),
					      bMonochrome ? fWaveLength : -nColorRays));
				
/*	double fAngle = fDirection + (nRays > 1 ? 
				      2*fWidth * i / (nRays-1) -fWidth :
				      0 );
	raysToTrace.addElement (new LightRay (fX,fY,
					      Math.cos (fAngle),
					      -Math.sin (fAngle),
					      bMonochrome ? fWaveLength : -nColorRays));
					      */
      }
  }

  public boolean IsSource ()
  {
    return true;
  }

  public Vector GetHandlePoints ()
  {
    Vector v = super.GetHandlePoints ();
    
    v.addElement (new Point ((int) (( fLength*10.0*
				     Math.cos (0.5*Math.PI+fDirection))),
			     (int) ((-fLength*10.0*
				     Math.sin (0.5*Math.PI+fDirection)))));
    return v;
  }
}
