import java.awt.*;
import java.util.*;
import java.lang.*;
import OpticsObject;
import OpticsObjPanel;
import LightRay;
import HitObject;
import HitMirror;

import java.lang.Math;

class PlaneMirrorObject extends OpticsObject
{
  protected static final double MAX_LENGTH = 10.0;

  double fLength = 5.0;

  static int nObjectNo = 1;
  public String GetUniqueName () { return "Plane Mirror "+(nObjectNo++); }
  public static String GetPanelName () { return "PlaneMirrorPanel"; }
  public static String GetObjectName () { return "Plane Mirror"; }

  public void SetPanel (OpticsObjPanel panel)
  {
    super.SetPanel (panel);
    if (thePanel != null)
      {
	((PlaneMirrorPanel) thePanel).SetLength (fLength);
      }
  }

  public void SetLength (double length,boolean setPanel)
  {
    theLabCanvas.AbortCalculation ();
    fLength = length;
    fRadius = fLength;
    nLabRadius = (int) (fLength * 10.0);
    if (setPanel && thePanel != null)
      ((PlaneMirrorPanel) thePanel).SetLength (fLength);
    theLabCanvas.repaint ();
    theLabCanvas.Recalculate ();
  }

  public void DrawContours (Graphics g,boolean selected)
  {
    g = InitDrawContours (g,selected);

    double fXLength =  fLength*10.0*Math.cos (0.5*Math.PI+fDirection);
    double fYLength = -fLength*10.0*Math.sin (0.5*Math.PI+fDirection);

    double fXOffset =  6.0*Math.cos (fDirection);
    double fYOffset = -6.0*Math.sin (fDirection);

    g.drawLine ((int) fXLength,(int) fYLength,
		(int) -fXLength,(int) -fYLength);
    g.drawLine ((int) (0.5*fXOffset+fXLength),(int) (0.5*fYOffset+fYLength),
		(int) (0.5*fXOffset-fXLength),(int) (0.5*fYOffset-fYLength));
    g.drawLine ((int) fXLength,(int) fYLength,
		(int) (fXOffset+fXLength),(int) (fYOffset+fYLength));
    g.drawLine ((int) -fXLength,(int) -fYLength,
		(int) (fXOffset-fXLength),(int) (fYOffset-fYLength));
  }

  public void HandlePointMove (int nPoint,Point pt)
  {
    super.HandlePointMove (nPoint,pt);

    if (nPoint == 1)
      {
	double fDirection = GetPointDirection (pt);
	double fLength = GetPointLength (pt);

	if (fLength > MAX_LENGTH)
	  fLength = MAX_LENGTH;
	SetDirection (ToPiInterval (fDirection-0.5*Math.PI),true);
	SetLength (fLength,true);
      }
  }
  
  public Vector GetHandlePoints ()
  {
    Vector v = super.GetHandlePoints ();
    
    v.addElement (new Point ((int) (( fLength*10.0*
				     Math.cos (0.5*Math.PI+fDirection))),
			     (int) ((-fLength*10.0*
				     Math.sin (0.5*Math.PI+fDirection)))));
    return v;
  }

  public double IsHit (LightRay ray)
  {
    return IntersectLine (ray,
			  fX + fLength*Math.cos (0.5*Math.PI+fDirection),
			  fY - fLength*Math.sin (0.5*Math.PI+fDirection),
			  (- Math.cos (0.5*Math.PI+fDirection)),
			  (+ Math.sin (0.5*Math.PI+fDirection)),
			  fLength,1);
  }

  public HitObject GetHit (LightRay ray)
  {
    return new HitMirror (IntersectHitLine (ray,
					    fX + fLength*
					    Math.cos (0.5*Math.PI+fDirection),
					    fY - fLength*
					    Math.sin (0.5*Math.PI+fDirection),
					    (- Math.cos (0.5*Math.PI+fDirection)),
					    (+ Math.sin (0.5*Math.PI+fDirection)),
					    1,Math.PI-fDirection));
  }
}
