import java.awt.*;
import java.lang.Math;
import java.util.*;

import LightSourceObject;
import PointLightSourcePanel;
import LightRay;

class PointLightSourceObject extends LightSourceObject
{
  double fWidth = 5*Math.PI/180.0;

  PointLightSourceObject ()
  {
    nLabRadius = 20;
  }

  static int nObjectNo = 1;
  public String GetUniqueName () { return "Point Source "+(nObjectNo++); }
  public static String GetPanelName () { return "PointLightSourcePanel"; }
  public static String GetObjectName () { return "Point Light Source"; }

  public void SetPanel (OpticsObjPanel panel)
  {
    super.SetPanel (panel);
    if (thePanel != null)
      ((PointLightSourcePanel) thePanel).SetWidth (fWidth);
  }
  
  public void SetWidth (double width,boolean setPanel)
  {
    theLabCanvas.AbortCalculation ();
    fWidth = width;
    if (setPanel && thePanel != null)
      ((PointLightSourcePanel) thePanel).SetWidth (fWidth);
    theLabCanvas.repaint ();
    theLabCanvas.Recalculate ();
  }

  public void DrawContours (Graphics g,boolean selected)
  {
    g = InitDrawContours (g,selected);

    g.drawArc (-20,-20,
	       40,40,
	       RadiansToIntDegrees(fDirection+fWidth),
	       360-RadiansToIntDegrees(2*fWidth));
    g.drawOval (-10,-10,
	       20,20);
    g.drawLine (-3,-3,3,3);
    g.drawLine (3,-3,-3,3);
  }

  public void HandlePointMove (int nPoint,Point pt)
  {
    super.HandlePointMove (nPoint,pt);

    if (nPoint == 1)
      {
	double fDirection = GetPointDirection (pt);

	SetDirection (fDirection,true);
      }
    if (nPoint == 2)
      {
	double fWidth = GetPointAngle (pt);

	fWidth = Math.abs (fWidth);

	SetWidth (fWidth,true);
      }
  }

  public void GetRays (Vector raysToTrace)
  { 
    for (int i = 0; i < nRays; i++)
      {
	double fAngle = fDirection + (nRays > 1 ? 
				      2*fWidth * i / (nRays-1) -fWidth :
				      0 );
	raysToTrace.addElement (new LightRay (fX,fY,
					      Math.cos (fAngle),
					      -Math.sin (fAngle),
					      bMonochrome ? fWaveLength : -nColorRays));
      }
  }

  public boolean IsSource ()
  {
    return true;
  }

  public Vector GetHandlePoints ()
  {
    Vector v = super.GetHandlePoints ();
    
    v.addElement (new Point ((int) (( 10.0*Math.cos (fDirection))),
			     (int) ((-10.0*Math.sin (fDirection)))));
    v.addElement (new Point ((int) (( 20.0*Math.cos (fDirection+fWidth))),
			     (int) ((-20.0*Math.sin (fDirection+fWidth)))));
    return v;
  }
}
