import java.lang.*;
import java.util.*;

import LabCanvas;
import LightRay;
import OpticsObject;
import HitPaper;
import HitMirror;
import HitGlass;

class RayCalculator extends Thread
{
  LabCanvas theBoss;

  Vector raysToTrace;
  Vector theObjects;

  OpticsObject theObjs[];

  RayCalculator (LabCanvas boss,Vector objects)
  {
    theBoss = boss;
    theObjects = objects;
  }

  public void run ()
  {
    // first, get all the rays we have to play with,
    // then trace them one at the time

    raysToTrace = new Vector ();

    int nNonSources = 0;

    OpticsObject obj;

    for (int i = 0; i < theObjects.size (); i++)
      {
	if ((obj = (OpticsObject) theObjects.elementAt (i)).IsSource ())
	  obj.GetRays (raysToTrace);
	else
	  nNonSources++;
	obj.NewTrace ();
      }
    yield ();
    theObjs = new OpticsObject[nNonSources];
    int j = 0;
    for (int i = 0; i < theObjects.size (); i++)
      if (!(obj = (OpticsObject) theObjects.elementAt (i)).IsSource ())
	theObjs[j++] = obj;

    for (int i = 0; i < raysToTrace.size (); i++)
      {
	LightRay ray = (LightRay) raysToTrace.elementAt (i);

	TraceRay (ray);
      }

    for (int i = 0; i < theObjects.size (); i++)
      ((OpticsObject) theObjects.elementAt (i)).TraceFinished ();

    LightRay theRays[] = new LightRay[raysToTrace.size ()];
    
    for (int i = 0; i < raysToTrace.size (); i++)
	theRays[i] = (LightRay) raysToTrace.elementAt (i);

    // now sort the rays by intensity

    QuickSortRays (theRays,0,theRays.length-1);

    for (int i = 0; i < theRays.length-1; i++)
	if (theRays[i].fIntensity > theRays[i+1].fIntensity)
	    System.out.println ("Messup...");

    theBoss.CalculationFinished (theRays);
  }

  void QuickSortRays (LightRay rays[],int first,int last)
    {
	int pivot, front, back;
	
	if (first >= last)
	    return;

	float fPivotIntensity = rays[first].fIntensity;
	
	front = first;
	back = last+1;
	
	for ( ; ; )
	    {
		do
		    {
			front++;
		    }
		while (fPivotIntensity > rays[front].fIntensity && front < last);
		do
		    {
			back--;
		    }
		while (rays[back].fIntensity > fPivotIntensity);
		if (front < back)
		    {
			LightRay tmp = rays[front]; 
			rays[front] = rays[back]; 
			rays[back] = tmp;
		    }
		else
		    break;
	    }
	LightRay tmp = rays[first]; 
	rays[first] = rays[back]; 
	rays[back] = tmp;
			
	if (first+1 < back)
		QuickSortRays(rays,first,back-1);
	if (back+1 < last)
		QuickSortRays(rays,back+1,last);
    }

  double ToPiInterval (double rad)
  {
    while (rad > Math.PI)
      rad -= 2*Math.PI;
    while (rad < -Math.PI)
      rad += 2*Math.PI;
    return rad;
  }

  void TraceRay (LightRay ray)
  {
    //    System.out.println ("ray ("+ray.pX+","+ray.pY+
    //			")  ->("+ray.vX+","+ray.vY+")");

    // find any blocking objects

    OpticsObject obj;
    OpticsObject objHit = null;
    double fClosestMatch = Double.MAX_VALUE;
    double fDist;
    for (int i = 0; i < theObjs.length; i++)
      {
	obj = theObjs[i];

	double bx = obj.fX-ray.pX;
	double by = obj.fY-ray.pY;

	if (Math.abs (bx*ray.vY-by*ray.vX) < obj.fRadius &&
	    Math.sqrt(bx*bx+by*by)-obj.fRadius < fClosestMatch)
	  {
	    // we are close enough to possibly be closer than anyone before
	    // and we are inside the objects enclosing circle

	    if ((fDist = obj.IsHit (ray)) < Double.MAX_VALUE &&
		fDist < fClosestMatch)
	      {
		fClosestMatch = fDist;
		objHit = obj;
	      }
	  }
      }

    if (objHit != null)
      {
	HitObject hit = objHit.GetHit (ray);
	if (ray.nBounces < 25)
	  {
	    if (hit instanceof HitPaper)
	      {
		  // already taken care of		
	      }
	    else if (hit instanceof HitMirror)
	      {
		HitMirror hitMirror = (HitMirror) hit;

		/*AddRayToTrace(new LightRay(hit.fX,hit.fY,
					   Math.cos (hitMirror.fNormalAngle),
					   Math.sin (hitMirror.fNormalAngle),1000,
					   objHit,hit.nSurface,1,1,600));*/
		
		double fRayAngle = Math.atan2(ray.vY,ray.vX);
		double fDiffAngle = ToPiInterval(Math.PI-
						 (hitMirror.fNormalAngle - 
						  fRayAngle));
		
		if (Math.abs(fDiffAngle) < 0.5*Math.PI)
		  {
		    double fAngle = hitMirror.fNormalAngle-fDiffAngle;
		    
		    AddRayToTrace (new LightRay (hit.fX,hit.fY,
						 Math.cos (fAngle),
						 Math.sin (fAngle),
						 ray.nBounces+1,
						 objHit,
						 hit.nSurface,
						 ray.fEfieldParallell,
						 ray.fEfieldPerpendicular,
						 ray.fWaveLength));
		  }	    
	      }
	    else if (hit instanceof HitGlass)
	      {
		HitGlass hitGlass = (HitGlass) hit;
		
		/*AddRayToTrace(new LightRay(hit.fX,hit.fY,
					   Math.cos (hitGlass.fNormalAngle),
					   Math.sin (hitGlass.fNormalAngle),1000,
					   objHit,hit.nSurface,1,1,600));*/

		double fRayAngle = Math.atan2(ray.vY,ray.vX);
		double fDiffAngle = ToPiInterval(Math.PI-
						 (hitGlass.fNormalAngle - 
						  fRayAngle));
		
		double fAngle = hitGlass.fNormalAngle-fDiffAngle;

		double fReflectedTotal[] = new double[2];
		
		if (ray.fWaveLength < 0)
		  {
		    int rays = (int) (-ray.fWaveLength+0.5);

		    double fReflected[] = new double[2];
		    int nTransmittedRays = 0;
		    fReflectedTotal[0] = fReflectedTotal[1] = 0;
	    
		    for (int i = 0; i < rays; i++)
		      {
			  if (AddRaySplitWavelengths (rays < 2 ? 
						      (float) ((LightRay.MAX_WAVE_LENGTH+
								LightRay.MIN_WAVE_LENGTH) * 0.5) : 
						      (float) (LightRay.MIN_WAVE_LENGTH+
							       (LightRay.MAX_WAVE_LENGTH-
								LightRay.MIN_WAVE_LENGTH)*i/(rays-1)),
						      fDiffAngle,fReflected,hitGlass,
						      ray,objHit) == 1)
			      {
				  nTransmittedRays++;
				  fReflectedTotal[0] += fReflected[0];
				  fReflectedTotal[1] += fReflected[1];
			      }
		      }
		    fReflectedTotal[0] = 
			(fReflectedTotal[0] + rays-nTransmittedRays)/rays;
		    fReflectedTotal[1] = 
			(fReflectedTotal[1] + rays-nTransmittedRays)/rays;

		    //System.out.println ("1: "+fReflectedTotal[0]+"  "+fReflectedTotal[1]);
		  }
		else
		    {
			AddRaySplitWavelengths (ray.fWaveLength,
						fDiffAngle,fReflectedTotal,
						hitGlass,ray,objHit);
			//System.out.println ("2: "+fReflectedTotal[0]+"  "+fReflectedTotal[1]);
		    }

		AddRayToTrace (new LightRay (hit.fX,hit.fY,
					     Math.cos (fAngle),
					     Math.sin (fAngle),
					     ray.nBounces+1,
					     objHit,
					     hit.nSurface,
					     (float) (ray.fEfieldParallell*
						      Math.sqrt(fReflectedTotal[0])),
					     (float) (ray.fEfieldPerpendicular*
						      Math.sqrt(fReflectedTotal[1])),
					     ray.fWaveLength));
		
	      }
	  }
	ray.vX = hit.fX;
	ray.vY = hit.fY;
	return;
      }

    // no blocking objects, terminate ray at -10000/10000 boundaries

    double tx = ((ray.vX > 0.0 ? 10000 : -10000)-ray.pX)/ray.vX;
    double ty = ((ray.vY > 0.0 ? 10000 : -10000)-ray.pY)/ray.vY;

    double t = Math.min(tx,ty);

    ray.vX = ray.pX + ray.vX * t;
    ray.vY = ray.pY + ray.vY * t;

    //    System.out.println ("ray ("+ray.pX+","+ray.pY+
    //			") -> ("+ray.vX+","+ray.vY+")");
  }

    void AddRayToTrace (LightRay ray)
    {
	// don't add it if it's to faint

	if (ray.fIntensity < 0.01)
	    return;
	raysToTrace.addElement (ray);
    }

  int AddRaySplitWavelengths (float fWaveLength,
			       double fDiffAngle,
			       double fReflected[],
			       HitGlass hitGlass,
			       LightRay ray,
			       OpticsObject objHit)
    {
      double fRefractionIndex = Math.sqrt (1+hitGlass.fRefractionConsts[0]/
					   (hitGlass.fRefractionConsts[1]-
					    1/(fWaveLength*fWaveLength)));
      
      double fTransAngle = 
	Math.abs(fDiffAngle) < 0.5*Math.PI ?
	Math.sin(fDiffAngle)/fRefractionIndex :
	Math.sin(fDiffAngle)*fRefractionIndex;

      fReflected[0] = 1;
      fReflected[1] = 1;
      
      if (Math.abs(fTransAngle) <= 1.0)
	{
	    fTransAngle = Math.asin (fTransAngle);

	    double fTransCalcAngle = ToZeroHalfPiInterval(fTransAngle);
	    double fDiffCalcAngle = ToZeroHalfPiInterval(fDiffAngle);

	    fReflected[0] = Math.pow(Math.abs(fDiffAngle+fTransAngle) < 1e-9 ?
		((fRefractionIndex-1)/(fRefractionIndex+1)) :
		(/*-*/Math.tan(fDiffCalcAngle-fTransCalcAngle)/
		Math.tan(fDiffCalcAngle+fTransCalcAngle)),2);
	    fReflected[1] = Math.pow(Math.abs(fDiffAngle+fTransAngle) < 1e-9 ?
		((fRefractionIndex-1)/(fRefractionIndex+1)) :
		(/*-*/Math.sin(fDiffCalcAngle-fTransCalcAngle)/
		Math.sin(fDiffCalcAngle+fTransCalcAngle)),2);

	    double fAngle = 
		Math.abs(fDiffAngle) < 0.5*Math.PI ?
		Math.PI+hitGlass.fNormalAngle+fTransAngle:
		hitGlass.fNormalAngle-fTransAngle;
	    
	    AddRayToTrace (new LightRay (hitGlass.fX,hitGlass.fY,
					 Math.cos (fAngle),
					 Math.sin (fAngle),
					 ray.nBounces+1,
					 objHit,
					 hitGlass.nSurface,
					 (float) (ray.fEfieldParallell*
						  Math.sqrt(1-fReflected[0])),
					 (float) (ray.fEfieldPerpendicular*
						  Math.sqrt(1-fReflected[1])),
					 fWaveLength));
	    return 1;
	}
      return 0;
    }

  double ToZeroHalfPiInterval (double rad)
    {
	rad = Math.abs(rad);
	if (rad > Math.PI/2)
	    rad = Math.PI - rad;
	return rad;
    }

}




